import logging
import time

from fuocore.models import cached_field
from fuocore.models import (
    BaseModel,
    SongModel,
    AlbumModel,
    ArtistModel,
    MvModel,
    PlaylistModel,
    LyricModel,
    SearchModel,
    UserModel,
    GeneratorProxy,
    SearchType,
)

from .provider import provider

logger = logging.getLogger(__name__)


class XBaseModel(BaseModel):
    _api = provider.api

    class Meta:
        allow_get = True
        provider = provider


def _deserialize(data, schema_cls):
    schema = schema_cls()
    obj = schema.load(data)
    return obj


def create_g(func, identifier, field='songs', schema=None):
    if schema is None:
        schema = NestedSongSchema
    data = func(identifier, page=1)
    # user_favorite_songs 接口返回的数据有 total 字段，
    # 但 playlist_detail_v2 接口返回的数据没有 total 字段，
    # 这里取 pagingVO 结构体中的 count 字段值作为 total
    total = int(data['pagingVO']['count'])

    def g():
        nonlocal data
        if data is None:
            yield from ()
        else:
            paging = data['pagingVO']
            # pagingVO 结构体中字段是 string 类型
            page = int(paging['page'])
            page_size = int(paging['pageSize'])
            pages = int(paging['pages'])
            while page <= pages:
                obj_data_list = data[field]
                for obj_data in obj_data_list:
                    yield _deserialize(obj_data, schema)
                page += 1
                data = func(identifier, page, page_size)
    # FIXME: use SequentialRandomReader instead of GeneratorProxy
    return GeneratorProxy(g(), total)


class XMvModel(MvModel, XBaseModel):
    @classmethod
    def get(cls, identifier):
        data = cls._api.mv_detail(identifier)
        if data is not None:
            return _deserialize(data, MvSchema)


class XSongModel(SongModel, XBaseModel):

    class Meta:
        fields = ['mvid', 'q_media_mapping', 'expired_at']
        fields_no_get = ['expired_at']
        support_multi_quality = True

    @classmethod
    def get(cls, identifier):
        data = cls._api.song_detail(identifier)
        if data is None:
            return None
        return _deserialize(data, SongSchema)

    def refresh_url(self):
        song = self.get(self.identifier)
        self.url = song.url
        self.q_media_mapping = song.q_media_mapping
        self.expired_at = song.expired_at

    @property
    def is_expired(self):
        return self.expired_at is not None and time.time() >= self.expired_at

    @property
    def url(self):
        if time.time() > self._expired_at:
            logger.debug('song({}) url is expired, refresh...'.format(self))
            self.refresh_url()
        return self._url

    @url.setter
    def url(self, value):
        self._expired_at = time.time() + 60 * 60 * 1  # one hour
        self._url = value

    @property
    def lyric(self):
        if self._lyric is not None:
            return self._lyric
        content = self._api.song_lyric(self.identifier)
        self.lyric = LyricModel(
            identifier=self.identifier,
            content=content
        )
        return self._lyric

    @lyric.setter
    def lyric(self, value):
        self._lyric = value

    @property
    def mv(self):
        if self._mv is not None:
            return self._mv
        # 这里可能会先获取一次 mvid
        if self.mvid:
            mv = XMvModel.get(self.mvid)
            if mv is not None:
                self._mv = mv
                return self._mv
        return None

    @mv.setter
    def mv(self, value):
        self._mv = value

    # multi quality support

    def list_quality(self):
        return list((self.q_media_mapping or {}).keys())

    def get_media(self, quality):
        if self.is_expired:
            self.refresh_url()
        return self.q_media_mapping.get(quality)


class XAlbumModel(AlbumModel, XBaseModel):

    @classmethod
    def get(cls, identifier):
        data = cls._api.album_detail(identifier)
        if data is None:
            return None
        return _deserialize(data, AlbumSchema)


class XArtistModel(ArtistModel, XBaseModel):
    class Meta:
        allow_create_songs_g = True
        allow_create_albums_g = True

    @classmethod
    def get(cls, identifier):
        data = cls._api.artist_detail(identifier)
        if data is None:
            return None
        return _deserialize(data, ArtistSchema)

    @property
    def songs(self):
        if self._songs is None:
            self._songs = []
            data_songs = self._api.artist_songs(self.identifier)['songs'] or []
            if data_songs:
                for data_song in data_songs:
                    song = _deserialize(data_song, NestedSongSchema)
                    self._songs.append(song)
        return self._songs

    def create_songs_g(self):
        return create_g(self._api.artist_songs, self.identifier)

    @songs.setter
    def songs(self, value):
        self._songs = value

    def create_albums_g(self):
        return create_g(self._api.artist_albums, self.identifier, 'albums', AlbumSchema)


class XPlaylistModel(PlaylistModel, XBaseModel):

    class Meta:
        fields = ('uid', )
        allow_create_songs_g = True

    @classmethod
    def get(cls, identifier):
        # FIXME: 获取所有歌曲，而不仅仅是前 100 首
        data = cls._api.playlist_detail(identifier)
        if data is None:
            return None
        return _deserialize(data, PlaylistSchema)

    def add(self, song_id, **kwargs):
        rv = self._api.update_playlist_song(self.identifier, song_id, 'add')
        if rv:
            song = XSongModel.get(song_id)
            self.songs.append(song)
            return True
        return rv

    def remove(self, song_id, allow_not_exist=True):
        rv = self._api.update_playlist_song(self.identifier, song_id, 'del')
        for song in self.songs:
            if song.identifier == song_id:
                self.songs.remove(song)
        return rv

    def create_songs_g(self):
        return create_g(self._api.playlist_detail_v2, self.identifier)


class XSearchModel(SearchModel, XBaseModel):
    pass


class XUserModel(UserModel, XBaseModel):
    class Meta:
        allow_fav_songs_add = True
        allow_fav_songs_remove = True
        fields = ('access_token', )
        fields_no_get = ('fav_songs', 'fav_albums', 'fav_artists',
                         'rec_songs', 'rec_playlists')

    @classmethod
    def get(cls, identifier):
        user_data = cls._api.user_detail(identifier)
        if user_data is None:
            return None
        return _deserialize(user_data, UserSchema)

    @cached_field()
    def playlists(self):
        """获取用户创建的歌单

        如果不是用户本人，则不能获取用户默认精选集
        """
        playlists_data = self._api.user_playlists(self.identifier)
        playlists = []
        for playlist_data in playlists_data:
            playlist = _deserialize(playlist_data, PlaylistSchema)
            playlists.append(playlist)
        return playlists

    @cached_field()
    def fav_playlists(self):
        playlists_data = self._api.user_favorite_playlists(self.identifier)
        fav_playlists = []
        for playlist_data in playlists_data:
            playlist = _deserialize(playlist_data, PlaylistSchema)
            fav_playlists.append(playlist)
        return fav_playlists

    @cached_field()
    def rec_playlists(self):
        playlists_data = self._api.recommend_playlists()
        rec_playlists = []
        for playlist_data in playlists_data:
            playlist = _deserialize(playlist_data, PlaylistSchema)
            rec_playlists.append(playlist)
        return rec_playlists

    @property
    def fav_songs(self):
        return create_g(self._api.user_favorite_songs, self.identifier)

    @fav_songs.setter
    def fav_songs(self, _): pass

    def add_to_fav_songs(self, song_id):
        return self._api.update_favorite_song(song_id, 'add')

    def remove_from_fav_songs(self, song_id):
        return self._api.update_favorite_song(song_id, 'del')

    @property
    def fav_artists(self):
        return create_g(self._api.user_favorite_artists,
                        self.identifier, 'artists', ArtistSchema)

    @fav_artists.setter
    def fav_artists(self, _): pass

    @property
    def fav_albums(self):
        return create_g(self._api.user_favorite_albums,
                        self.identifier, 'albums', AlbumSchema)

    @fav_albums.setter
    def fav_albums(self, _): pass

    @cached_field()
    def rec_songs(self):
        songs_data = self._api.recommend_songs()
        return [_deserialize(song_data, SongSchema)
                for song_data in songs_data]

    def get_radio(self):
        songs_data = self._api.personal_fm()
        if songs_data is None:
            logger.error('data should not be None')
            return None
        return [_deserialize(song_data, SongSchema)
                for song_data in songs_data]


def search(keyword, **kwargs):
    type_ = SearchType.parse(kwargs['type_'])
    type_type_map = {
        SearchType.so: 1,
        SearchType.al: 10,
        SearchType.ar: 100,
        SearchType.pl: 1000,
    }
    data = provider.api.search(keyword, type_=type_type_map[type_])
    result = _deserialize(data, SearchSchema)
    result.q = keyword
    return result


from .schemas import (  # noqa
    AlbumSchema,
    ArtistSchema,
    MvSchema,
    PlaylistSchema,
    NestedSongSchema,
    SongSchema,
    SearchSchema,
    UserSchema,
)



import logging
import time
from urllib.parse import urlparse

from marshmallow import Schema, fields, post_load, EXCLUDE

from fuocore.media import Media


logger = logging.getLogger(__name__)


class ArtistSchema(Schema):
    """歌手详情 Schema、歌曲歌手简要信息 Schema
    """

    class Meta:
        unknown = EXCLUDE

    identifier = fields.Int(data_key='artistId', required=True)
    name = fields.Str(data_key='artistName', required=True)
    cover = fields.Str(data_key='artistLogo', missing=None)
    desc = fields.Str(data_key='description', missing=None)

    @post_load
    def create_model(self, data, **kwargs):
        return XArtistModel(**data)


class ListenFileSchema(Schema):
    """Song listenfile"""

    class Meta:
        unknown = EXCLUDE

    quality = fields.Str(required=True)
    # song_detail and artist_songs api return different listenFile struct,
    # the struct song_detail api return only contains url field, while
    # the struct artist_songs api return only contains listenFile field.
    url = fields.Str(missing=None)
    url_bak = fields.Str(data_key='listenFile', missing=None)
    format = fields.Str(required=True)
    # expire = fields.Str(required=True)

    @post_load
    def process_url(self, data, **kwargs):
        url = data['url'] or data['url_bak']
        data['url'] = url
        return data

    @classmethod
    def to_q_media_mapping(cls, lfiles):
        q_media_mapping = {}
        if lfiles:
            q_q_mapping = {'m': 'shq',  # for example: flac
                           's': 'shq',
                           'h': 'hq',
                           'l': 'sq',
                           'f': 'lq',
                           'e': 'lq'}
            for lfile in filter(lambda lfile: lfile['url'], lfiles):
                url = lfile['url']
                quality = lfile['quality']
                format = lfile['format']
                # url example: http://m720.xiami.net/...
                try:
                    bitrate = int(urlparse(url).netloc.split('.')[0][1:])
                except:  # noqa
                    bitrate = None
                if quality not in q_q_mapping:
                    field = 'lq'
                    logger.warning('unknown quality {}, url {}'.format(quality, url))
                else:
                    field = q_q_mapping[quality]
                q_media_mapping[field] = Media(url, format=format, bitrate=bitrate)
        return q_media_mapping


class AlbumSchema(Schema):
    """专辑详情 Schema

    >>> import json
    >>> with open('data/fixtures/xiami/album.json') as f:
    ...     data = json.load(f)
    ...     schema = AlbumSchema(strict=True)
    ...     album, _ = schema.load(data)
    >>> album.identifier
    2100387382
    """
    class Meta:
        unknown = EXCLUDE

    identifier = fields.Int(data_key='albumId', required=True)
    name = fields.Str(data_key='albumName', required=True)
    cover = fields.Str(data_key='albumLogo', required=True)

    songs = fields.List(fields.Nested('NestedSongSchema'))
    artists = fields.List(fields.Nested(ArtistSchema), data_key='artists')
    desc = fields.Str(data_key='description')

    @post_load
    def create_model(self, data, **kwargs):
        return XAlbumModel(**data)


class MvSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    identifier = fields.Str(requried=True, data_key='mvId')
    name = fields.Str(requried=True, data_key='title')
    cover = fields.Str(requried=True, data_key='mvCover')
    media = fields.Str(requried=True, data_key='mp4Url')

    @post_load
    def create_model(self, data, **kwargs):
        return XMvModel(**data)


class SongSchema(Schema):
    """歌曲详情 Schema

    >>> import json
    >>> with open('data/fixtures/xiami/song.json') as f:
    ...     data = json.load(f)
    ...     schema = SongSchema(strict=True)
    ...     song, _ = schema.load(data)
    >>> song.url
    ''
    """
    class Meta:
        unknown = EXCLUDE

    identifier = fields.Int(data_key='songId', required=True)
    mvid = fields.Str(requried=True, data_key='mvId')
    title = fields.Str(data_key='songName', required=True)
    # FIXME: 有的歌曲没有 length 字段
    duration = fields.Str(data_key='length', missing='0')

    url = fields.Str(data_key='listenFile', missing='')
    files = fields.List(
        fields.Nested(ListenFileSchema), data_key='listenFiles', missing=[])

    # XXX: 这里暂时用 singerVOs 来表示歌曲的 artist，即使虾米接口中
    # 也会包含歌曲 artistVOs 信息
    artists = fields.List(
            fields.Nested(ArtistSchema), data_key='singerVOs', required=True)

    album_id = fields.Int(data_key='albumId', required=True)
    album_name = fields.Str(data_key='albumName', required=True)
    album_cover = fields.Str(data_key='albumLogo', required=True)

    @post_load
    def create_model(self, data, **kwargs):
        album = XAlbumModel(identifier=data['album_id'],
                            name=data['album_name'],
                            cover=data['album_cover'])
        files = data['files']
        if files:
            url = files[0]['url']
        else:
            url = ''
        q_media_mapping = ListenFileSchema.to_q_media_mapping(files)
        expire = int(time.time()) + 60 * 60
        song = XSongModel(identifier=data['identifier'],
                          mvid=data['mvid'],
                          title=data['title'],
                          url=url,
                          duration=int(data['duration']),
                          album=album,
                          artists=data['artists'],
                          q_media_mapping=q_media_mapping,
                          expired_at=expire,)
        return song


class NestedSongSchema(SongSchema):
    """搜索结果中歌曲 Schema、专辑/歌手详情中歌曲 Schema

    通过 search 得到的 Song 的结构和通过 song_detail 获取的 Song 的结构不一样

    search 接口得到的 Song 没有 listenFile 字段，但是可能会有 listenFiles 字段，
    有的话，取 listenFiles 中最高质量的播放链接作为音乐的 url。
    """
    class Meta:
        unknown = EXCLUDE

    @post_load
    def create_model(self, data, **kwargs):
        song = super().create_model(data)
        files = data['files']
        if files:
            song.url = files[0]['url']
        return song


class PlaylistSchema(Schema):
    """歌单 Schema

    >>> import json
    >>> with open('data/fixtures/xiami/playlist.json') as f:
    ...     data = json.load(f)
    ...     schema = PlaylistSchema(strict=True)
    ...     playlist, _ = schema.load(data)
    >>> len(playlist.songs)
    100
    """
    class Meta:
        unknown = EXCLUDE

    identifier = fields.Str(data_key='listId', required=True)
    uid = fields.Int(data_key='userId', required=True)
    name = fields.Str(data_key='collectName', required=True)
    cover = fields.Str(data_key='collectLogo', required=True)
    songs = fields.List(fields.Nested(NestedSongSchema), missing=None)
    desc = fields.Str(data_key='description', missing=None)

    @post_load
    def create_model(self, data, **kwargs):
        return XPlaylistModel(**data)


class SearchSchema(Schema):
    """搜索结果 Schema"""
    class Meta:
        unknown = EXCLUDE

    songs = fields.List(fields.Nested(NestedSongSchema))
    albums = fields.List(fields.Nested(AlbumSchema))
    artists = fields.List(fields.Nested(ArtistSchema))
    playlists = fields.List(fields.Nested(PlaylistSchema), data_key='collects')

    @post_load
    def create_model(self, data, **kwargs):
        return XSearchModel(**data)


class UserSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    identifier = fields.Int(data_key='userId')
    name = fields.Str(data_key='nickName')
    access_token = fields.Str(data_key='accessToken')

    @post_load
    def create_model(self, data, **kwargs):
        return XUserModel(**data)


from .models import (  # noqa
    XAlbumModel,
    XArtistModel,
    XMvModel,
    XPlaylistModel,
    XSongModel,
    XSearchModel,
    XUserModel,
)
